#include <pgsql/libpq-fe.h>
#include <emodules.h>

/*
 * This is an interface to libpq, the C library interface to
 * PostgreSQL. I have attempted a direct 1:1 mapping of C to lisp.
 *
 * The current version is very incomplete and only implements the
 * functions I needed for a very simple calendar program.
 */

struct pg_conn
{
  struct lcrecord_header header;
  PGconn *conn;
};

DECLARE_EXTERNAL_LRECORD (pg_conn, struct pg_conn);
#define XPG_CONN(x) XRECORD (x, pg_conn, struct pg_conn)
#define XSETPG_CONN(x, p) XSETRECORD (x, p, pg_conn)
#define PG_CONNP(x) RECORDP (x, pg_conn)
#define CHECK_PG_CONN(x) CHECK_RECORD (x, pg_conn)
#define CONCHECK_PG_CONN(x) CONCHECK_RECORD (x, pg_conn)
#define XREALPG_CONN(x) XPG_CONN(x)->conn

Lisp_Object Qpg_connp;

static void
print_pg_conn (Lisp_Object obj, Lisp_Object printcharfun, int escapeflag)
{
  char buf[40];

  sprintf (buf, "#<%s 0x%08lx>",
	   XRECORD_LHEADER_IMPLEMENTATION (obj)->name,
	   (unsigned long) XREALPG_CONN (obj));
  write_c_string (buf, printcharfun);
}

static Lisp_Object
make_pg_conn (PGconn *conn_ptr)
{
  struct pg_conn *conn;
  Lisp_Object lisp_res;

  conn = alloc_lcrecord_type (struct pg_conn, &lrecord_pg_conn);
  conn->conn = conn_ptr;
  XSETPG_CONN (lisp_res, conn);
  return lisp_res;
}

DEFINE_EXTERNAL_LRECORD_IMPLEMENTATION ("pg-conn", pg_conn, 0,
                                        print_pg_conn, 0, 0, 0, 0,
                                        struct pg_conn);

struct pg_result
{
  struct lcrecord_header header;
  PGresult *res;
};

DECLARE_EXTERNAL_LRECORD (pg_result, struct pg_result);
#define XPG_RESULT(x) XRECORD (x, pg_result, struct pg_result)
#define XSETPG_RESULT(x, p) XSETRECORD (x, p, pg_result)
#define PG_RESULTP(x) RECORDP (x, pg_result)
#define CHECK_PG_RESULT(x) CHECK_RECORD (x, pg_result)
#define CONCHECK_PG_RESULT(x) CONCHECK_RECORD (x, pg_result)
#define XREALPG_RESULT(x) XPG_RESULT(x)->res

Lisp_Object Qpg_resultp;

static void
print_pg_result (Lisp_Object obj, Lisp_Object printcharfun, int escapeflag)
{
  char buf[40];

  sprintf (buf, "#<%s 0x%08lx>",
	   XRECORD_LHEADER_IMPLEMENTATION (obj)->name,
	   (unsigned long) XREALPG_RESULT (obj));
  write_c_string (buf, printcharfun);
}

static Lisp_Object
make_pg_result (PGresult *res_ptr)
{
  struct pg_result *result;
  Lisp_Object lisp_res;

  result = alloc_lcrecord_type (struct pg_result, &lrecord_pg_result);
  result->res = res_ptr;
  XSETPG_RESULT (lisp_res, result);
  return lisp_res;
}

DEFINE_EXTERNAL_LRECORD_IMPLEMENTATION ("pg-result", pg_result, 0,
                                        print_pg_result, 0, 0, 0, 0,
                                        struct pg_result);

DEFUN ("PQconnectdb", FPQconnectdb, 1, 1, "", /*
Connect to the PostgreSQL database.
Returns a connection or nil.
*/
	   (arg_string))
{
	CHECK_STRING(arg_string);
	return make_pg_conn(PQconnectdb(XSTRING_DATA(arg_string)));
}

DEFUN ("PQfinish", FPQfinish, 1, 1, "", /*
Disconnect from the SQL database.
*/
	   (conn))
{
	CHECK_PG_CONN(conn);
	PQfinish(XREALPG_CONN(conn));
	return Qt;
}


DEFUN ("PQexec", FPQexec, 2, 2, "", /*
Exec call.
Returns a results index or nil.
*/
	   (conn, command))
{
	CHECK_PG_CONN(conn);
	CHECK_STRING(command);
	return make_pg_result(PQexec(XREALPG_CONN(conn),
								 XSTRING_DATA(command)));
}


DEFUN ("PQclear", FPQclear, 1, 1, "", /*
Free up a result.
*/
	   (result))
{
	CHECK_PG_RESULT(result);
	PQclear(XREALPG_RESULT(result));
	return Qt;
}


DEFUN ("PQresultStatus", FPQresultStatus, 1, 1, "", /*
Return the result status as an integer.
*/
	   (result))
{
	CHECK_PG_RESULT(result);
	return make_int(PQresultStatus(XREALPG_RESULT(result)));
}


DEFUN ("PQresStatus", FPQresStatus, 1, 1, "", /*
Given a result status as an integer, return a status string.
*/
	   (rc))
{
	CHECK_INT(rc);
	return build_string(PQresStatus(XINT(rc)));
}


DEFUN ("PQresultErrorMessage", FPQresultErrorMessage, 1, 1, "", /*
Return the result status as a string.
*/
	   (result))
{
	CHECK_PG_RESULT(result);
	return build_string(PQresultErrorMessage(XREALPG_RESULT(result)));
}


DEFUN ("PQntuples", FPQntuples, 1, 1, "", /*
Return number of tuples in query result.
*/
	   (result))
{
	CHECK_PG_RESULT(result);
	return make_int(PQntuples(XREALPG_RESULT(result)));
}


DEFUN ("PQbinaryTuples", FPQbinaryTuples, 1, 1, "", /*
Return number of binary tuples in query result.
*/
	   (result))
{
	CHECK_PG_RESULT(result);
	return make_int(PQbinaryTuples(XREALPG_RESULT(result)));
}


DEFUN ("PQnfields", FPQnfields, 1, 1, "", /*
Return number of fields in query result.
*/
	   (result))
{
	CHECK_PG_RESULT(result);
	return make_int(PQnfields(XREALPG_RESULT(result)));
}


DEFUN ("PQgetvalue", FPQgetvalue, 3, 3, "", /*
Get a value result.
*/
	   (result, tuple, field))
{
	CHECK_PG_RESULT(result);
	CHECK_INT(tuple);
	CHECK_INT(field);
	return build_string(PQgetvalue(XREALPG_RESULT(result),
								   XINT(tuple), XINT(field)));
}


void
syms_of_PostgreSQL()
{
	INIT_EXTERNAL_LRECORD_IMPLEMENTATION (pg_conn);
	INIT_EXTERNAL_LRECORD_IMPLEMENTATION (pg_result);

	DEFSYMBOL(Qpg_connp);
	DEFSYMBOL(Qpg_resultp);

	DEFSUBR(FPQconnectdb);
	DEFSUBR(FPQfinish);
	DEFSUBR(FPQexec);
	DEFSUBR(FPQclear);
	DEFSUBR(FPQresultStatus);
	DEFSUBR(FPQresStatus);
	DEFSUBR(FPQresultErrorMessage);
	DEFSUBR(FPQntuples);
	DEFSUBR(FPQbinaryTuples);
	DEFSUBR(FPQnfields);
	DEFSUBR(FPQgetvalue);
}

void
vars_of_PostgreSQL()
{
}
